﻿using System;
using System.IO;

namespace TestLibrary
{
    /// <summary>
    /// C# 코드 작성자
    /// </summary>
    public class CSharpCodeWriter : ICodeWriter
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// NO RENAME ATTRIBUTE
        /// </summary>
        private const string NO_RENAME_ATTRIBUTE = "[Obfuscation(Feature = \"renaming\", Exclude = true)]";

        /// <summary>
        /// NO PRUNE ATTRIBUTE
        /// </summary>
        private const string NO_PRUNE_ATTRIBUTE = "[Obfuscation(Feature = \"trigger\", Exclude = false)]";

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (ICodeWriter) 파일 확장자 - FileExtension

        /// <summary>
        /// 파일 확장자
        /// </summary>
        public string FileExtension
        {
            get
            {
                return ".cs";
            }
        }

        #endregion
        #region (ICodeWriter) 디스플레이명 - DisplayName

        /// <summary>
        /// 디스플레이명
        /// </summary>
        public string DisplayName
        {
            get
            {
                return "C#";
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (ICodeWriter) 타입명 구하기 - GetTypeName(type, generator)

        /// <summary>
        /// 타입명 구하기
        /// </summary>
        /// <param name="type">타입</param>
        /// <param name="generator">제너레이터</param>
        /// <returns>타입명</returns>
        public string GetTypeName(JSONType type, IJSONClassGenerator generator)
        {
            bool arraysAsLists = !generator.ExplicitDeserialization;

            switch(type.ValueType)
            {
                case JSONValueType.Anything          : return "object";
                case JSONValueType.Array             : return arraysAsLists ? "IList<" + GetTypeName(type.InternalType, generator) + ">" :
                                                                              GetTypeName(type.InternalType, generator) + "[]";
                case JSONValueType.Dictionary        : return "Dictionary<string, " + GetTypeName(type.InternalType, generator) + ">";
                case JSONValueType.Boolean           : return "bool";
                case JSONValueType.Float             : return "double";
                case JSONValueType.Integer           : return "int";
                case JSONValueType.Long              : return "long";
                case JSONValueType.Date              : return "DateTime";
                case JSONValueType.NonConstrained    : return "object";
                case JSONValueType.NullableBoolean   : return "bool?";
                case JSONValueType.NullableFloat     : return "double?";
                case JSONValueType.NullableInteger   : return "int?";
                case JSONValueType.NullableLong      : return "long?";
                case JSONValueType.NullableDate      : return "DateTime?";
                case JSONValueType.NullableSomething : return "object";
                case JSONValueType.Object            : return type.AssignedName;
                case JSONValueType.String            : return "string";
                default                              : throw new NotSupportedException("Unsupported json type");
            }
        }

        #endregion
        #region (ICodeWriter) 클래스 쓰기 - WriteClass(generator, writer, type)

        /// <summary>
        /// 클래스 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        public void WriteClass(IJSONClassGenerator generator, TextWriter writer, JSONType type)
        {
            string visibility = generator.InternalVisibility ? "internal" : "public";

            if(generator.UseNestedClasses)
            {
                if(!type.IsRoot)
                {
                    if(ShouldApplyNoRenamingAttribute(generator))
                    {
                        writer.WriteLine("        " + NO_RENAME_ATTRIBUTE);
                    }

                    if(ShouldApplyNoPruneAttribute(generator))
                    {
                        writer.WriteLine("        " + NO_PRUNE_ATTRIBUTE);
                    }

                    writer.WriteLine("        {0} class {1}", visibility, type.AssignedName);
                    writer.WriteLine("        {");
                }
            }
            else
            {
                if(ShouldApplyNoRenamingAttribute(generator))
                {
                    writer.WriteLine("    " + NO_RENAME_ATTRIBUTE);
                }

                if(ShouldApplyNoPruneAttribute(generator))
                {
                    writer.WriteLine("    " + NO_PRUNE_ATTRIBUTE);
                }

                writer.WriteLine("    {0} class {1}", visibility, type.AssignedName);
                writer.WriteLine("    {");
            }

            string prefix = generator.UseNestedClasses && !type.IsRoot ? "            " : "        ";

            bool shouldSuppressWarning = generator.InternalVisibility && !generator.UseProperties && !generator.ExplicitDeserialization;

            if(shouldSuppressWarning)
            {
                writer.WriteLine("#pragma warning disable 0649");

                if(!generator.UsePascalCase)
                {
                    writer.WriteLine();
                }
            }

            if(type.IsRoot && generator.ExplicitDeserialization)
            {
                WriteStringWithConstructorExplicitDeserialization(generator, writer, type, prefix);
            }

            if(generator.ExplicitDeserialization)
            {
                if(generator.UseProperties)
                {
                    WriteClassWithPropertiesExplicitDeserialization(writer, type, prefix);
                }
                else
                {
                    WriteClassWithFieldsExplicitDeserialization(writer, type, prefix);
                }
            }
            else
            {
                WriteClassMembers(generator, writer, type, prefix);
            }

            if(shouldSuppressWarning)
            {
                writer.WriteLine();
                writer.WriteLine("#pragma warning restore 0649");
                writer.WriteLine();
            }

            if(generator.UseNestedClasses && !type.IsRoot)
            {
                writer.WriteLine("        }");
            }

            if(!generator.UseNestedClasses)
            {
                writer.WriteLine("    }");
            }

            writer.WriteLine();
        }

        #endregion
        #region (ICodeWriter) 파일 시작 쓰기 - WriteFileStart(generator, writer)

        /// <summary>
        /// 파일 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        public void WriteFileStart(IJSONClassGenerator generator, TextWriter writer)
        {
            if(generator.UseNamespaces)
            {
                foreach(string line in JSONClassGenerator.FileHeaderArray)
                {
                    writer.WriteLine("// " + line);
                }

                writer.WriteLine();

                writer.WriteLine("using System;");
                writer.WriteLine("using System.Collections.Generic;");

                if(ShouldApplyNoPruneAttribute(generator) || ShouldApplyNoRenamingAttribute(generator))
                {
                    writer.WriteLine("using System.Reflection;");
                }

                if(!generator.ExplicitDeserialization && generator.UsePascalCase)
                {
                    writer.WriteLine("using Newtonsoft.Json;");
                }

                writer.WriteLine("using Newtonsoft.Json.Linq;");

                if(generator.ExplicitDeserialization)
                {
                    writer.WriteLine("using TestLibrary;");
                }

                if(generator.SecondaryNamespace != null && generator.HasSecondaryClasses && !generator.UseNestedClasses)
                {
                    writer.WriteLine("using {0};", generator.SecondaryNamespace);
                }
            }

            if(generator.UseNestedClasses)
            {
                writer.WriteLine("    {0} class {1}", generator.InternalVisibility ? "internal" : "public", generator.MainClass);
                writer.WriteLine("    {");
            }
        }

        #endregion
        #region (ICodeWriter) 파일 종료 쓰기 - WriteFileEnd(generator, writer)

        /// <summary>
        /// 파일 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        public void WriteFileEnd(IJSONClassGenerator generator, TextWriter writer)
        {
            if(generator.UseNestedClasses)
            {
                writer.WriteLine("    }");
            }
        }

        #endregion
        #region (ICodeWriter) 네임스페이스 시작 쓰기 - WriteNamespaceStart(generator, writer, root)

        /// <summary>
        /// 네임스페이스 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        public void WriteNamespaceStart(IJSONClassGenerator generator, TextWriter writer, bool root)
        {
            writer.WriteLine();
            writer.WriteLine("namespace {0}", root && !generator.UseNestedClasses ? generator.Namespace : (generator.SecondaryNamespace ?? generator.Namespace));
            writer.WriteLine("{");
            writer.WriteLine();
        }

        #endregion
        #region (ICodeWriter) 네임스페이스 종료 쓰기 - WriteNamespaceEnd(generator, writer, root)

        /// <summary>
        /// 네임스페이스 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        public void WriteNamespaceEnd(IJSONClassGenerator generator, TextWriter writer, bool root)
        {
            writer.WriteLine("}");
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 이름 변경 어트리뷰트 미적용 의무 여부 구하기 - ShouldApplyNoRenamingAttribute(generator)

        /// <summary>
        /// 이름 변경 어트리뷰트 미적용 의무 여부 구하기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <returns>이름 변경 어트리뷰트 미적용 의무 여부</returns>
        private bool ShouldApplyNoRenamingAttribute(IJSONClassGenerator generator)
        {
            return generator.ApplyObfuscationAttributes && !generator.ExplicitDeserialization && !generator.UsePascalCase;
        }

        #endregion
        #region 정리 어트리뷰트 미적용 여부 구하기 - ShouldApplyNoPruneAttribute(generator)

        /// <summary>
        /// 정리 어트리뷰트 미적용 여부 구하기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <returns>정리 어트리뷰트 미적용 여부</returns>
        private bool ShouldApplyNoPruneAttribute(IJSONClassGenerator generator)
        {
            return generator.ApplyObfuscationAttributes && !generator.ExplicitDeserialization && generator.UseProperties;
        }

        #endregion
        #region 클래스 멤버 쓰기 - WriteClassMembers(generator, writer, type, prefix)

        /// <summary>
        /// 클래스 멤버 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        /// <param name="prefix">접두사</param>
        private void WriteClassMembers(IJSONClassGenerator generator, TextWriter writer, JSONType type, string prefix)
        {
            foreach(FieldInfo fieldInfo in type.FieldInfoList)
            {
                if(generator.UsePascalCase || generator.ExamplesInDocumentation)
                {
                    writer.WriteLine();
                }

                if(generator.ExamplesInDocumentation)
                {
                    writer.WriteLine(prefix + "/// <summary>");
                    writer.WriteLine(prefix + "/// Examples: " + fieldInfo.GetExampleText());
                    writer.WriteLine(prefix + "/// </summary>");
                }

                if(generator.UsePascalCase)
                {
                    writer.WriteLine(prefix + "[JsonProperty(\"{0}\")]", fieldInfo.JSONMemberName);
                }

                if(generator.UseProperties)
                {
                    writer.WriteLine(prefix + "public {0} {1} {{ get; set; }}", fieldInfo.Type.GetTypeName(), fieldInfo.MemberName);
                }
                else
                {
                    writer.WriteLine(prefix + "public {0} {1};", fieldInfo.Type.GetTypeName(), fieldInfo.MemberName);
                }
            }
        }

        #endregion

        #region 명시적 역직렬화 속성를 갖는 클래스 쓰기 - WriteClassWithPropertiesExplicitDeserialization(writer, type, prefix)

        /// <summary>
        /// 명시적 역직렬화 속성를 갖는 클래스 쓰기
        /// </summary>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        /// <param name="prefix">접두사</param>
        private void WriteClassWithPropertiesExplicitDeserialization(TextWriter writer, JSONType type, string prefix)
        {
            writer.WriteLine(prefix + "private JObject __jobject;");
            writer.WriteLine(prefix + "public {0}(JObject obj)", type.AssignedName);
            writer.WriteLine(prefix + "{");
            writer.WriteLine(prefix + "    this.__jobject = obj;");
            writer.WriteLine(prefix + "}");
            writer.WriteLine();

            foreach(FieldInfo fieldInfo in type.FieldInfoList)
            {
                string variable = null;

                if(fieldInfo.Type.MustCache)
                {
                    variable = "_" + char.ToLower(fieldInfo.MemberName[0]) + fieldInfo.MemberName.Substring(1);

                    writer.WriteLine(prefix + "[System.Diagnostics.DebuggerBrowsable(System.Diagnostics.DebuggerBrowsableState.Never)]");
                    writer.WriteLine(prefix + "private {0} {1};", fieldInfo.Type.GetTypeName(), variable);
                }

                writer.WriteLine(prefix + "public {0} {1}", fieldInfo.Type.GetTypeName(), fieldInfo.MemberName);
                writer.WriteLine(prefix + "{");
                writer.WriteLine(prefix + "    get");
                writer.WriteLine(prefix + "    {");

                if(fieldInfo.Type.MustCache)
                {
                    writer.WriteLine(prefix + "        if ({0} == null)", variable);
                    writer.WriteLine(prefix + "            {0} = {1};", variable, fieldInfo.GetGenerationCode("__jobject"));
                    writer.WriteLine(prefix + "        return {0};", variable);
                }
                else
                {
                    writer.WriteLine(prefix + "        return {0};", fieldInfo.GetGenerationCode("__jobject"));
                }

                writer.WriteLine(prefix + "    }");
                writer.WriteLine(prefix + "}");
                writer.WriteLine();
            }
        }

        #endregion
        #region 명시적 역직렬화 생성자를 갖는 문자열 쓰기 - WriteStringWithConstructorExplicitDeserialization(generator, writer, type, prefix)

        /// <summary>
        /// 명시적 역직렬화 생성자를 갖는 문자열 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        /// <param name="prefix">접두사</param>
        private void WriteStringWithConstructorExplicitDeserialization(IJSONClassGenerator generator, TextWriter writer, JSONType type, string prefix)
        {
            writer.WriteLine();
            writer.WriteLine(prefix + "public {1}(string json)", generator.InternalVisibility ? "internal" : "public", type.AssignedName);
            writer.WriteLine(prefix + "    : this(JObject.Parse(json))");
            writer.WriteLine(prefix + "{");
            writer.WriteLine(prefix + "}");
            writer.WriteLine();
        }

        #endregion
        #region 명시적 역직렬화 필드를 갖는 클래스 쓰기 - WriteClassWithFieldsExplicitDeserialization(writer, type, prefix)

        /// <summary>
        /// 명시적 역직렬화 필드를 갖는 클래스 쓰기
        /// </summary>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        /// <param name="prefix">접두사</param>
        private void WriteClassWithFieldsExplicitDeserialization(TextWriter writer, JSONType type, string prefix)
        {
            writer.WriteLine(prefix + "public {0}(JObject obj)", type.AssignedName);
            writer.WriteLine(prefix + "{");

            foreach(FieldInfo fieldInfo in type.FieldInfoList)
            {
                writer.WriteLine(prefix + "    this.{0} = {1};", fieldInfo.MemberName, fieldInfo.GetGenerationCode("obj"));
            }

            writer.WriteLine(prefix + "}");
            writer.WriteLine();

            foreach(FieldInfo fieldInfo in type.FieldInfoList)
            {
                writer.WriteLine(prefix + "public readonly {0} {1};", fieldInfo.Type.GetTypeName(), fieldInfo.MemberName);
            }
        }

        #endregion
    }
}