﻿using System;
using System.IO;

namespace TestLibrary
{
    /// <summary>
    /// 타입 스크립트 코드 작성자
    /// </summary>
    public class TypeScriptCodeWriter : ICodeWriter
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (ICodeWriter) 파일 확장자 - FileExtension

        /// <summary>
        /// 파일 확장자
        /// </summary>
        public string FileExtension
        {
            get
            {
                return ".ts";
            }
        }

        #endregion
        #region (ICodeWriter) 디스플레이명 - DisplayName

        /// <summary>
        /// 디스플레이명
        /// </summary>
        public string DisplayName
        {
            get
            {
                return "TypeScript";
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (ICodeWriter) 타입명 구하기 - GetTypeName(type, generator)

        /// <summary>
        /// 타입명 구하기
        /// </summary>
        /// <param name="type">타입</param>
        /// <param name="generator">제너레이터</param>
        /// <returns>타입명</returns>
        public string GetTypeName(JSONType type, IJSONClassGenerator generator)
        {
            switch(type.ValueType)
            {
                case JSONValueType.Anything          : return "any";
                case JSONValueType.String            : return "string";
                case JSONValueType.Boolean           : return "bool";
                case JSONValueType.Integer           :
                case JSONValueType.Long              :
                case JSONValueType.Float             : return "number";
                case JSONValueType.Date              : return "Date";
                case JSONValueType.NullableInteger   :
                case JSONValueType.NullableLong      :
                case JSONValueType.NullableFloat     : return "number";
                case JSONValueType.NullableBoolean   : return "bool";
                case JSONValueType.NullableDate      : return "Date";
                case JSONValueType.Object            : return type.AssignedName;
                case JSONValueType.Array             : return GetTypeName(type.InternalType, generator) + "[]";
                case JSONValueType.Dictionary        : return "{ [key: string]: " + GetTypeName(type.InternalType, generator) + "; }";
                case JSONValueType.NullableSomething : return "any";
                case JSONValueType.NonConstrained    : return "any";
                default                              : throw new NotSupportedException("Unsupported type");
            }
        }

        #endregion
        #region (ICodeWriter) 클래스 쓰기 - WriteClass(generator, writer, type)

        /// <summary>
        /// 클래스 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        public void WriteClass(IJSONClassGenerator generator, TextWriter writer, JSONType type)
        {
            string prefix = GetNamespace(generator, type.IsRoot) != null ? "    " : "";

            bool exported = !generator.InternalVisibility || generator.SecondaryNamespace != null;

            writer.WriteLine(prefix + (exported ? "export " : string.Empty) + "interface " + type.AssignedName + " {");

            foreach(FieldInfo fieldInfo in type.FieldInfoList)
            {
                bool shouldDefineNamespace = type.IsRoot                          &&
                                             generator.SecondaryNamespace != null &&
                                             generator.Namespace          != null &&
                                             (
                                                 fieldInfo.Type.ValueType == JSONValueType.Object ||
                                                 (fieldInfo.Type.InternalType != null && fieldInfo.Type.InternalType.ValueType == JSONValueType.Object)
                                             );

                if(generator.ExamplesInDocumentation)
                {
                    writer.WriteLine();
                    writer.WriteLine(prefix + "    /**");
                    writer.WriteLine(prefix + "      * Examples: " + fieldInfo.GetExampleText());
                    writer.WriteLine(prefix + "      */");
                }

                writer.WriteLine
                (
                    prefix                                                                      +
                    "    "                                                                      +
                    fieldInfo.JSONMemberName                                                    +
                    (IsNullable(fieldInfo.Type.ValueType) ? "?" : "")                           +
                    ": "                                                                        +
                    (shouldDefineNamespace ? generator.SecondaryNamespace + "." : string.Empty) +
                    GetTypeName(fieldInfo.Type, generator)                                      +
                    ";"
                );
            }

            writer.WriteLine(prefix + "}");
            writer.WriteLine();
        }

        #endregion
        #region (ICodeWriter) 파일 시작 쓰기 - WriteFileStart(generator, writer)

        /// <summary>
        /// 파일 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        public void WriteFileStart(IJSONClassGenerator generator, TextWriter writer)
        {
            foreach(string line in JSONClassGenerator.FileHeaderArray)
            {
                writer.WriteLine("// " + line);
            }

            writer.WriteLine();
        }

        #endregion
        #region (ICodeWriter) 파일 종료 쓰기 - WriteFileEnd(generator, writer)

        /// <summary>
        /// 파일 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        public void WriteFileEnd(IJSONClassGenerator generator, TextWriter writer)
        {
        }

        #endregion
        #region (ICodeWriter) 네임스페이스 시작 쓰기 - WriteNamespaceStart(generator, writer, root)

        /// <summary>
        /// 네임스페이스 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        public void WriteNamespaceStart(IJSONClassGenerator generator, TextWriter writer, bool root)
        {
            if(GetNamespace(generator, root) != null)
            {
                writer.WriteLine("module " + GetNamespace(generator, root) + " {");
                writer.WriteLine();
            }
        }

        #endregion
        #region (ICodeWriter) 네임스페이스 종료 쓰기 - WriteNamespaceEnd(generator, writer, root)

        /// <summary>
        /// 네임스페이스 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        public void WriteNamespaceEnd(IJSONClassGenerator generator, TextWriter writer, bool root)
        {
            if(GetNamespace(generator, root) != null)
            {
                writer.WriteLine("}");
                writer.WriteLine();
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region NULL 가능 여부 구하기 - IsNullable(valueType)

        /// <summary>
        /// NULL 가능 여부 구하기
        /// </summary>
        /// <param name="valueType">값 타입</param>
        /// <returns>NULL 가능 여부</returns>
        private bool IsNullable(JSONValueType valueType)
        {
            return valueType == JSONValueType.NullableBoolean ||
                   valueType == JSONValueType.NullableDate    ||
                   valueType == JSONValueType.NullableFloat   ||
                   valueType == JSONValueType.NullableInteger ||
                   valueType == JSONValueType.NullableLong    ||
                   valueType == JSONValueType.NullableSomething;
        }

        #endregion
        #region 네임스페이스 구하기 - GetNamespace(generator, root)

        /// <summary>
        /// 네임스페이스 구하기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="root">루트 여부</param>
        /// <returns>네임스페이스</returns>
        private string GetNamespace(IJSONClassGenerator generator, bool root)
        {
            return root ? generator.Namespace : (generator.SecondaryNamespace ?? generator.Namespace);
        }

        #endregion
    }
}