﻿using System;
using System.IO;

namespace TestLibrary
{
    /// <summary>
    /// 비주얼 베이직 코드 작성자
    /// </summary>
    public class VisualBasicCodeWriter : ICodeWriter
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// NO RENAME ATTRIBUTE
        /// </summary>
        private const string NO_RENAME_ATTRIBUTE = "<Obfuscation(Feature:=\"renaming\", Exclude:=true)>";

        /// <summary>
        /// NO PRUNE ATTRIBUTE
        /// </summary>
        private const string NO_PRUNE_ATTRIBUTE = "<Obfuscation(Feature:=\"trigger\", Exclude:=false)>";

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (ICodeWriter) 파일 확장자 - FileExtension

        /// <summary>
        /// 파일 확장자
        /// </summary>
        public string FileExtension
        {
            get
            {
                return ".vb";
            }
        }

        #endregion
        #region (ICodeWriter) 디스플레이명 - DisplayName

        /// <summary>
        /// 디스플레이명
        /// </summary>
        public string DisplayName
        {
            get
            {
                return "Visual Basic .NET";
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (ICodeWriter) 타입명 구하기 - GetTypeName(type, generator)

        /// <summary>
        /// 타입명 구하기
        /// </summary>
        /// <param name="type">타입</param>
        /// <param name="generator">제너레이터</param>
        /// <returns>타입명</returns>
        public string GetTypeName(JSONType type, IJSONClassGenerator generator)
        {
            bool arraysAsLists = generator.ExplicitDeserialization;

            switch(type.ValueType)
            {
                case JSONValueType.Anything          : return "Object";
                case JSONValueType.Array             : return arraysAsLists ? "IList(Of " + GetTypeName(type.InternalType, generator) + ")" :
                                                                              GetTypeName(type.InternalType, generator) + "()";
                case JSONValueType.Dictionary        : return "Dictionary(Of String, " + GetTypeName(type.InternalType, generator) + ")";
                case JSONValueType.Boolean           : return "Boolean";
                case JSONValueType.Float             : return "Double";
                case JSONValueType.Integer           : return "Integer";
                case JSONValueType.Long              : return "Long";
                case JSONValueType.Date              : return "DateTime";
                case JSONValueType.NonConstrained    : return "Object";
                case JSONValueType.NullableBoolean   : return "Boolean?";
                case JSONValueType.NullableFloat     : return "Double?";
                case JSONValueType.NullableInteger   : return "Integer?";
                case JSONValueType.NullableLong      : return "Long?";
                case JSONValueType.NullableDate      : return "DateTime?";
                case JSONValueType.NullableSomething : return "Object";
                case JSONValueType.Object            : return type.AssignedName;
                case JSONValueType.String            : return "String";
                default                              : throw new NotSupportedException("Unsupported json type");
            }
        }

        #endregion
        #region (ICodeWriter) 클래스 쓰기 - WriteClass(generator, writer, type)

        /// <summary>
        /// 클래스 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        public void WriteClass(IJSONClassGenerator generator, TextWriter writer, JSONType type)
        {
            string visibility = generator.InternalVisibility ? "Friend" : "Public";

            if(generator.UseNestedClasses)
            {
                writer.WriteLine("    {0} Partial Class {1}", visibility, generator.MainClass);

                if(!type.IsRoot)
                {
                    if(ShouldApplyNoRenamingAttribute(generator))
                    {
                        writer.WriteLine("        " + NO_RENAME_ATTRIBUTE);
                    }

                    if(ShouldApplyNoPruneAttribute(generator))
                    {
                        writer.WriteLine("        " + NO_PRUNE_ATTRIBUTE);
                    }

                    writer.WriteLine("        {0} Class {1}", visibility, type.AssignedName);
                }
            }
            else
            {
                if(ShouldApplyNoRenamingAttribute(generator))
                {
                    writer.WriteLine("    " + NO_RENAME_ATTRIBUTE);
                }

                if(ShouldApplyNoPruneAttribute(generator))
                {
                    writer.WriteLine("    " + NO_PRUNE_ATTRIBUTE);
                }

                writer.WriteLine("    {0} Class {1}", visibility, type.AssignedName);
            }

            string prefix = generator.UseNestedClasses && !type.IsRoot ? "            " : "        ";

            WriteClassMembers(generator, writer, type, prefix);

            if(generator.UseNestedClasses && !type.IsRoot)
            {
                writer.WriteLine("        End Class");
            }

            writer.WriteLine("    End Class");
            writer.WriteLine();
        }

        #endregion
        #region (ICodeWriter) 파일 시작 쓰기 - WriteFileStart(generator, writer)

        /// <summary>
        /// 파일 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        public void WriteFileStart(IJSONClassGenerator generator, TextWriter writer)
        {
            foreach(string line in JSONClassGenerator.FileHeaderArray)
            {
                writer.WriteLine("' " + line);
            }

            writer.WriteLine();
            writer.WriteLine("Imports System");
            writer.WriteLine("Imports System.Collections.Generic");

            if(ShouldApplyNoRenamingAttribute(generator) || ShouldApplyNoPruneAttribute(generator))
            {
                writer.WriteLine("Imports System.Reflection");
            }

            if(generator.UsePascalCase)
            {
                writer.WriteLine("Imports Newtonsoft.Json");
            }

            writer.WriteLine("Imports Newtonsoft.Json.Linq");

            if(generator.SecondaryNamespace != null && generator.HasSecondaryClasses && !generator.UseNestedClasses)
            {
                writer.WriteLine("Imports {0}", generator.SecondaryNamespace);
            }
        }

        #endregion
        #region (ICodeWriter) 파일 종료 쓰기 - WriteFileEnd(generator, writer)

        /// <summary>
        /// 파일 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        public void WriteFileEnd(IJSONClassGenerator generator, TextWriter writer)
        {
        }

        #endregion
        #region (ICodeWriter) 네임스페이스 시작 쓰기 - WriteNamespaceStart(generator, writer, root)

        /// <summary>
        /// 네임스페이스 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        public void WriteNamespaceStart(IJSONClassGenerator generator, TextWriter writer, bool root)
        {
            writer.WriteLine();
            writer.WriteLine("Namespace Global.{0}", root && !generator.UseNestedClasses ? generator.Namespace : (generator.SecondaryNamespace ?? generator.Namespace));
            writer.WriteLine();
        }

        #endregion
        #region (ICodeWriter) 네임스페이스 종료 쓰기 - WriteNamespaceEnd(generator, writer, root)

        /// <summary>
        /// 네임스페이스 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        public void WriteNamespaceEnd(IJSONClassGenerator generator, TextWriter writer, bool root)
        {
            writer.WriteLine("End Namespace");
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 이름 변경 어트리뷰트 미적용 의무 여부 구하기 - ShouldApplyNoRenamingAttribute(generator)

        /// <summary>
        /// 이름 변경 어트리뷰트 미적용 의무 여부 구하기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <returns>이름 변경 어트리뷰트 미적용 의무 여부</returns>
        private bool ShouldApplyNoRenamingAttribute(IJSONClassGenerator generator)
        {
            return generator.ApplyObfuscationAttributes && !generator.ExplicitDeserialization && !generator.UsePascalCase;
        }

        #endregion
        #region 정리 어트리뷰트 미적용 의무 여부 구하기 - ShouldApplyNoPruneAttribute(generator)

        /// <summary>
        /// 정리 어트리뷰트 미적용 의무 여부 구하기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <returns>정리 어트리뷰트 미적용 의무 여부</returns>
        private bool ShouldApplyNoPruneAttribute(IJSONClassGenerator generator)
        {
            return generator.ApplyObfuscationAttributes && !generator.ExplicitDeserialization && generator.UseProperties;
        }

        #endregion
        #region 클래스 멤버 쓰기 - WriteClassMembers(generator, writer, type, prefix)

        /// <summary>
        /// 클래스 멤버 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        /// <param name="prefix">접두사</param>
        private void WriteClassMembers(IJSONClassGenerator generator, TextWriter writer, JSONType type, string prefix)
        {
            foreach(FieldInfo fieldInfo in type.FieldInfoList)
            {
                if(generator.UsePascalCase || generator.ExamplesInDocumentation)
                {
                    writer.WriteLine();
                }

                if(generator.ExamplesInDocumentation)
                {
                    writer.WriteLine(prefix + "''' <summary>");
                    writer.WriteLine(prefix + "''' Examples: " + fieldInfo.GetExampleText());
                    writer.WriteLine(prefix + "''' </summary>");
                }

                if(generator.UsePascalCase)
                {
                    writer.WriteLine(prefix + "<JsonProperty(\"{0}\")>", fieldInfo.JSONMemberName);
                }

                if(generator.UseProperties)
                {
                    writer.WriteLine(prefix + "Public Property {1} As {0}", fieldInfo.Type.GetTypeName(), fieldInfo.MemberName);
                }
                else
                {
                    writer.WriteLine(prefix + "Public {1} As {0}", fieldInfo.Type.GetTypeName(), fieldInfo.MemberName);
                }
            }
        }

        #endregion
    }
}