﻿using System;
using System.Collections.Generic;

using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace TestLibrary
{
    /// <summary>
    /// JSON 클래스 헬퍼
    /// </summary>
    internal static class JSONClassHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Delegate
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 값 리더 대리자 - ValueReader<T>(token)

        /// <summary>
        /// 값 리더 대리자
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="token">토큰</param>
        /// <returns>값</returns>
        public delegate T ValueReader<T>(JToken token);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region JSON 토큰 구하기 - GetJToken<T>(jObject, field)

        /// <summary>
        /// JSON 토큰 구하기
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="jObject">JSON 객체</param>
        /// <param name="field">필드</param>
        /// <returns>JSON 토큰</returns>
        public static T GetJToken<T>(JObject jObject, string field) where T : JToken
        {
            JToken token;

            if(jObject.TryGetValue(field, out token))
            {
                return GetJToken<T>(token);
            }
            else
            {
                return null;
            }
        }

        #endregion
        #region 문자열 읽기 - ReadString(token)

        /// <summary>
        /// 문자열 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>문자열</returns>
        public static string ReadString(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                return null;
            }

            return (string)value.Value;
        }

        #endregion
        #region 진리값 읽기 - ReadBoolean(token)

        /// <summary>
        /// 진리값 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>진리값</returns>
        public static bool ReadBoolean(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                throw new JsonSerializationException();
            }

            return Convert.ToBoolean(value.Value);
        }

        #endregion
        #region NULL 가능 진리값 읽기 - ReadNullableBoolean(token)

        /// <summary>
        /// NULL 가능 진리값 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>NULL 가능 진리값</returns>
        public static bool? ReadNullableBoolean(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                return null;
            }

            return Convert.ToBoolean(value.Value);
        }

        #endregion
        #region 정수 읽기 - ReadInteger(token)

        /// <summary>
        /// 정수 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>정수</returns>
        public static int ReadInteger(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                throw new JsonSerializationException();
            }

            return Convert.ToInt32((long)value.Value);
        }

        #endregion
        #region NULL 가능 정수 읽기 - ReadNullableInteger(token)

        /// <summary>
        /// NULL 가능 정수 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>NULL 가능 정수</returns>
        public static int? ReadNullableInteger(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                return null;
            }

            return Convert.ToInt32((long)value.Value);
        }

        #endregion
        #region 배정도 정수 읽기 - ReadLong(token)

        /// <summary>
        /// 배정도 정수 읽기
        /// </summary>
        /// <param name="token"></param>
        /// <returns>배정도 정수</returns>
        public static long ReadLong(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                throw new JsonSerializationException();
            }

            return Convert.ToInt64(value.Value);
        }

        #endregion
        #region NULL 가능 배정도 정수 읽기 - ReadNullableLong(token)

        /// <summary>
        /// NULL 가능 배정도 정수 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>NULL 가능 배정도 정수</returns>
        public static long? ReadNullableLong(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                return null;
            }

            return Convert.ToInt64(value.Value);
        }

        #endregion
        #region 부동형 실수 읽기 - ReadFloat(token)

        /// <summary>
        /// 부동형 실수 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>부동형 실수</returns>
        public static double ReadFloat(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                throw new JsonSerializationException();
            }

            return Convert.ToDouble(value.Value);
        }

        #endregion
        #region NULL 가능 부동형 실수 읽기 - ReadNullableFloat(token)

        /// <summary>
        /// NULL 가능 부동형 실수 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>NULL 가능 부동형 실수</returns>
        public static double? ReadNullableFloat(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                return null;
            }

            return Convert.ToDouble(value.Value);
        }

        #endregion
        #region 날짜 읽기 - ReadDate(token)

        /// <summary>
        /// 날짜 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>날짜</returns>
        public static DateTime ReadDate(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                throw new JsonSerializationException();
            }

            return Convert.ToDateTime(value.Value);
        }

        #endregion
        #region NULL 가능 날짜 읽기 - ReadNullableDate(token)

        /// <summary>
        /// NULL 가능 날짜 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>NULL 가능 날짜</returns>
        public static DateTime? ReadNullableDate(JToken token)
        {
            JValue value = GetJToken<JValue>(token);

            if(value == null)
            {
                return null;
            }

            return Convert.ToDateTime(value.Value);
        }

        #endregion
        #region 객체 읽기 - ReadObject(token)

        /// <summary>
        /// 객체 읽기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <returns>객체</returns>
        public static object ReadObject(JToken token)
        {
            JToken value = GetJToken<JToken>(token);

            if(value == null)
            {
                return null;
            }

            if(value.Type == JTokenType.Object)
            {
                return value;
            }

            if(value.Type == JTokenType.Array)
            {
                return ReadArray<object>(value, ReadObject);
            }

            JValue jValue = value as JValue;

            if(jValue != null)
            {
                return jValue.Value;
            }

            return value;
        }

        #endregion
        #region 강력한 타입 객체 읽기 - ReadStronglyTypedObject<T>(token)

        /// <summary>
        /// 강력한 타입 객체 읽기
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="token">토큰</param>
        /// <returns>강력한 타입 객체</returns>
        public static T ReadStronglyTypedObject<T>(JToken token) where T : class
        {
            JObject value = GetJToken<JObject>(token);

            if(value == null)
            {
                return null;
            }

            return (T)Activator.CreateInstance(typeof(T), new object[] { token });
        }

        #endregion
        #region 배열 읽기 - ReadArray<T>(token, reader)

        /// <summary>
        /// 배열 읽기
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="token">토큰</param>
        /// <param name="reader">리더</param>
        /// <returns>배열</returns>
        public static T[] ReadArray<T>(JToken token, ValueReader<T> reader)
        {
            JArray value = GetJToken<JArray>(token);

            if(value == null)
            {
                return null;
            }

            T[] array = new T[value.Count];

            for(int i = 0; i < array.Length; i++)
            {
                array[i] = reader(value[i]);
            }

            return array;
        }

        #endregion
        #region 딕셔너리 읽기 - ReadDictionary<T>(token)

        /// <summary>
        /// 딕셔너리 읽기
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="token">토큰</param>
        /// <returns>딕셔너리</returns>
        public static Dictionary<string, T> ReadDictionary<T>(JToken token)
        {
            JObject value = GetJToken<JObject>(token);

            if(value == null)
            {
                return null;
            }
  
            Dictionary<string, T> dictionary = new Dictionary<string, T>();

            return dictionary;
        }

        #endregion
        #region 배열 읽기 - ReadArray<K>(jArray, reader, type)

        /// <summary>
        /// 배열 읽기
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="jArray">JSON 배열</param>
        /// <param name="reader">리더</param>
        /// <param name="type">타입</param>
        /// <returns>배열</returns>
        public static Array ReadArray<T>(JArray jArray, ValueReader<T> reader, Type type)
        {
            if(jArray == null)
            {
                return null;
            }

            Type elementType = type.GetElementType();

            Array array = Array.CreateInstance(elementType, jArray.Count);

            for(int i = 0; i < array.Length; i++)
            {
                if(elementType.IsArray)
                {
                    array.SetValue(ReadArray<T>(GetJToken<JArray>(jArray[i]), reader, elementType), i);
                }
                else
                {
                    array.SetValue(reader(jArray[i]), i);
                }
            }

            return array;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region JSON 토큰 구하기 - GetJToken<T>(token)

        /// <summary>
        /// JSON 토큰 구하기
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="token">토큰</param>
        /// <returns>JSON 토큰</returns>
        private static T GetJToken<T>(JToken token) where T : JToken
        {
            if(token == null)
            {
                return null;
            }

            if(token.Type == JTokenType.Null)
            {
                return null;
            }

            if(token.Type == JTokenType.Undefined)
            {
                return null;
            }

            return (T)token;
        }

        #endregion
    }
}